# ----------------------------------------------------------------------------
# Copyright (c) 2016-2022, QIIME 2 development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from .hooks import load_jupyter_server_extension
from .template import make_html

__all__ = ['make_html', 'load_jupyter_server_extension']
